/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * ----------------------------------------------------------------------
 * enhanced by Aley Keprt, 01d.05m.1998
 * (new F-key command set, US keyboard mapping, other keyboard extensions)
 * ----------------------------------------------------------------------
 *
 * File events.c : Event handling code - DOS version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "resource.h"
#include "util.h"
#include "version.h"
#include "mouse.h"
#include "spectrum.h"
#include "intrface.h"

#include "video.h"
#include "keyb_int.h"
#include "mous_int.h"
#include "audio.h"

static int EmulatorControlEvent = False;

int ctrlkey; //flag used for handling Ctrl+F1,F2,F3,etc.
int * key_table;
unsigned long pixels[NCOLOURS];

/* Maps vgalib scancodes to a Sam keyboard row and bit to set/reset */
/* E.g. ktbl[SCANCODE_A] means set/reset bit 0 (0xfe) of keyboard row 1 */
struct ktbl {
	int port;       /* which keyboard row is the key on? */
	int bit;        /* which bit to mask */
} keytable[51] = {
	{9, 0}, {9, 0}, 
	/* 1 2 3 4 5 6 7 8 9 0 */
	{3, 0xfe}, {3, 0xfd}, {3, 0xfb}, {3, 0xf7}, {3, 0xef}, 
	{4, 0xef}, {4, 0xf7}, {4, 0xfb}, {4, 0xfd}, {4, 0xfe},
			
	{9, 0}, {9, 0}, {9, 0}, {9,0},

	/* Q W E R T Y U I O P */
	{2, 0xfe}, {2, 0xfd}, {2, 0xfb}, {2, 0xf7}, {2, 0xef},
	{5, 0xef}, {5, 0xf7}, {5, 0xfb}, {5, 0xfd}, {5, 0xfe},
	
	{9, 0}, {9, 0}, {9, 0}, {9,0},

	/* A S D F G H J K L */
	{1, 0xfe}, {1, 0xfd}, {1, 0xfb}, {1, 0xf7}, {1, 0xef},
	{6, 0xef}, {6, 0xf7}, {6, 0xfb}, {6, 0xfd},
	
	{9, 0}, {9, 0}, {9, 0}, {9,0}, {9,0},

	/* Z X C V B N M */
		   {0, 0xfd}, {0, 0xfb}, {0, 0xf7}, {0, 0xef},
	{7, 0xef}, {7, 0xf7}, {7, 0xfb}
			
};

static struct k_map {
	char *res_name;
	int  Scancode;
} map_tab[NUM_KEYMAPPINGS] = {
	{ "JoystUp",    SCANCODE_9 }, 
	{ "JoystDown",  SCANCODE_8 },
	{ "JoystLeft",  SCANCODE_6 },
	{ "JoystRight", SCANCODE_7 },
	{ "JoystFire",  SCANCODE_0 },
};

uns8 keyports[10];      /* 8 rows of keys (+ 1 row for unscanned keys) */
uns8 inPorts[256], outPorts[256];

/* All to do with screen painting */

uns8 *vga_data_ptr;                     /* pointer to actual frame buffer data */               

extern int ClutChangeInFrame;           /* signals a clut change has been made in 
					   the present frame */
extern int BorderChangeInFrame;         /* signals a border change has been made in
					   the present frame */
extern int VmprChangeInFrame;           /* signals a vmpr change has been made in
					   the present frame */

int SpectrumKeyboard;                   /* Switch for spectrum keyboard
					   mode - see vgaspectrum.c */

static int VgaModeAvailable[2];         /* Flags for the display modes */
static int ScaleFactor;                 /* Mode for dynamic scale switching */

int Vga2x2LinearAddressing;

int Video1x1Mode;
int Video2x2Mode;
						  
static void SamKeyboardHandler(int Scancode, int EventType);
static void SamMouseHandler(int Button, int dx, int dy);
static void EmulatorControl(void);

extern int flash_phase;
extern uns8 LineNeedsUpdating[];

extern void (*line_update)(int);
extern void (*line_update_mode_1)(int);
extern void (*line_update_mode_2)(int);
extern void (*line_update_mode_3)(int);
extern void (*line_update_mode_4)(int);

void call_vmpr(uns8 val);


#define PointerButtonsOff  0xff		/* Default Mouse button status (Off) */
int SAMPointerMode = 1;                 /* Pointer Attatched and x magnification */

/* from processor.c */
extern int LineNo;
extern int disass;
extern int InterruptType;
 
/* The SAM ports */

uns8 vmpr = 0x81;
uns8 hmpr = 0x01;
uns8 lmpr = 0x1f;
uns8 midi = 0x00;       
uns8 border = 0x00;
uns8 clut[16] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 
		 0x00, 0x19, 0x2a, 0x3b, 0x4c, 0x5d, 0x6e, 0x7f};
uns8 keyboard = 0x5f;
uns8 status_reg = 0xff;
uns8 line_int = 0xff;

uns8 hpen = 0x00;
uns8 lpen = 0x00;
		
uns8 clock_reg = 0xff;

uns8 lepr = 0x00;
uns8 hepr = 0x00;

uns8 attr_reg = 0xff;


/* Pointer to the SAA Registers (NULL = No audio installed) */
uns8 *SAARegPtr = NULL;

/*
 *  Private functions start here
 */

/* Perform a clean close-down. */

static void
closedown(void)
{
	MouseClose();
	KeyboardClose();
	VideoFreeFramePointer();
	VideoClose();
}

static int ShiftPressed = False;                /* Flag for shift key */

/* Process a KeyPress event */
static void
process_keypress(int ks)
{
	/* First check for the 'special keys' */
	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] |= 0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] |= 0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] |= 0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] |= 0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] |= 0x01;

	switch (ks) {

		/* Ordinary SAM keys */

		case SCANCODE_SPACE:
			keyports[7] &= 0xfe; break;
		case SCANCODE_TAB:
			keyports[3] &= 0xbf; break;
		case SCANCODE_ENTER:
			keyports[6] &= 0xfe; break;
		case SCANCODE_LEFTCONTROL: case SCANCODE_RIGHTCONTROL:
			ctrlkey=1; //for handling Ctrl+F1,F2,F3,etc.
		/* <Shift> <Control> is mapped to Caps Lock */
			if (keyports[0] & 0x01)
			     keyports[8] &= 0xfe;
			else
			     keyports[3] &= 0x7f;
			break;
		case SCANCODE_LEFTALT:
/*              case SCANCODE_LEFTMETA: ????? */
			keyports[7] &= 0xfd; break;

		case SCANCODE_RIGHTALT: 
/*              case SCANCODE_RIGHTMETA: case SCANCODE_Mode_switch: ????? */
			keyports[6] &= 0x7f; break;
		case SCANCODE_BACKSPACE: 
/*		case SCANCODE_DELETE: ????? */
			keyports[4] &= 0x7f; break;
		case SCANCODE_ESCAPE:
			keyports[3] &= 0xdf; break;
		case SCANCODE_CURSORBLOCKUP:
			keyports[8] &= 0xfd; break;
		case SCANCODE_CURSORBLOCKDOWN:
			keyports[8] &= 0xfb; break;
		case SCANCODE_CURSORBLOCKLEFT:
			keyports[8] &= 0xf7; break;
		case SCANCODE_CURSORBLOCKRIGHT:
			keyports[8] &= 0xef; break;

		case SCANCODE_MINUS:
			if (!ShiftPressed)   
				keyports[4] &= 0xdf;
			else /* SCANCODE_UNDERSCORE */
				keyports[5] &= 0xdf;
			break;

		case SCANCODE_KEYPADDIVIDE:
			keyports[0] &= 0xfe; 
			keyports[4] &= 0xdf;
			break;

		case SCANCODE_KEYPADMULTIPLY:
			keyports[0] &= 0xfe; 
			keyports[4] &= 0xbf; 
			break;

		case SCANCODE_KEYPADMINUS:
			keyports[4] &= 0xdf;
			break;

		case SCANCODE_KEYPADPLUS:
			keyports[4] &= 0xbf;
			break;			

		case SCANCODE_EQUAL:
			if (!ShiftPressed)   
				keyports[5] &= 0xdf;
			else { /* SCANCODE_PLUS */
				keyports[0] |= 0x01; 
				keyports[4] &= 0xbf;
			}
			break;
		case SCANCODE_SEMICOLON:
			if (!ShiftPressed)   
				keyports[6] &= 0xdf;
			else { /* SCANCODE_COLON */
				keyports[0] |= 0x01; 
				keyports[6] &= 0xbf;
			}
			break;
		case SCANCODE_COMMA:
			if (!ShiftPressed) 
				keyports[7] &= 0xdf;
			else { /* SCANCODE_LESS */
				keyports[0] |= 0x01; keyports[7] &= 0xfd;
				keyports[2] &= 0xfe; 
			}
			break;  
		case SCANCODE_PERIOD:
			if (!ShiftPressed) 
				keyports[7] &= 0xbf;
			else { /* SCANCODE_GREATER */
				keyports[0] |= 0x01; keyports[7] &= 0xfd;
				keyports[2] &= 0xfd;
			}
			break;
		case SCANCODE_SLASH:
			if (!ShiftPressed) {
				keyports[0] &= 0xfe; 
				keyports[4] &= 0xdf;
			}
			else { /* SCANCODE_QUESTION */
				keyports[0] |= 0x01; keyports[7] &= 0xfd;
				keyports[0] &= 0xfb;
			}
			break;
		case SCANCODE_UK_BACKSLASH:
			if (!ShiftPressed) {
				keyports[0] &= 0xfe; keyports[7] &= 0x7f;
			}
			else {  /* SCANCODE_BAR */      
				keyports[0] |= 0x01; keyports[7] &= 0xfd;
				keyports[4] &= 0xfd;
			}
			break;
		case SCANCODE_7:
			if (!ShiftPressed) 
				keyports[4] &= 0xf7;
			else /* SCANCODE_AMPERSAND */                
				keyports[4] &=0xef; 
			break;
		case SCANCODE_BRACKET_LEFT:
			if (!ShiftPressed) {
				keyports[7] &= 0xfd; keyports[2] &= 0xf7;
			}
			else { /* SCANCODE_BRACERIGHT */
				keyports[0] |= 0x01;
				keyports[7] &= 0xfd; keyports[1] &= 0xf7;
			}
			break;
		case SCANCODE_BRACKET_RIGHT:
			if (!ShiftPressed) {
				keyports[7] &= 0xfd; keyports[2] &= 0xef;
			}
			else { /* SCANCODE_BRACERIGHT */
				keyports[0] |= 0x01;
				keyports[7] &= 0xfd; keyports[1] &= 0xef;
			}
			break;

		case SCANCODE_2:
			if (!ShiftPressed) 
				keyports[3] &= 0xfd;
			else { /* SCANCODE_QUOTEDBL */
#ifdef US_KEYBOARD
				keyports[3] &= 0xfd;
#else
				keyports[0] |= 0x01; 
				keyports[5] &= 0xbf;
#endif
			}                       
			break;
		case SCANCODE_APOSTROPHE:
			if (!ShiftPressed) { 
				keyports[0] &= 0xfe; 
				keyports[4] &= 0xf7;
			}
			else { /* SCANCODE_AT */
#ifdef US_KEYBOARD
				keyports[0] |= 0x01; 
				keyports[5] &= 0xbf;
#else
				keyports[3] &= 0xfd;
#endif
			}
			break;

#ifdef US_KEYBOARD
		case SCANCODE_BACKSLASH:
			if (!ShiftPressed) {
				//[\]
				keyports[0] &= 0xfe;
				keyports[7] &= 0x7f;
			} else {
				//[|] (Symbol+9)
				keyports[0] |= 0x01;
				keyports[7] &= 0xfd;
				keyports[4] &= 0xfd;
			}
			break;
#else
		case SCANCODE_UK_HASH:   /* NB Not in vgakeyboard.h */
			if (!ShiftPressed) {
				//hash
				keyports[0] &= 0xfe; 
				keyports[3] &= 0xfb;
			} else {
				//[~]
				keyports[4] &= 0xfe;
			}
			break;
#endif

#ifdef US_KEYBOARD
		case SCANCODE_GRAVE:
			if(!ShiftPressed) {
				//[`] = sterling
				keyports[7] &= 0xfd;
				keyports[6] &= 0xfd;
			}
			else {
				//tilde
				keyports[4] &= 0xfe;
			}
			break;
#endif

		case SCANCODE_6:
			if (!ShiftPressed) {
				keyports[4] &= 0xef;
			}
			else { /* SCANCODE_ASCIICIRCUM */
				keyports[0] |= 0x01; 
				keyports[7] &= 0xfd;
				keyports[6] &= 0xef;
			}
			break;

		case SCANCODE_3:
			if (!ShiftPressed) {
				keyports[3] &= 0xfb;
			}
			else { /* SCANCODE_STERLING */
#ifdef US_KEYBOARD
				keyports[3] &= 0xfb;
#else
				//Shift+3 = sterling
				keyports[0] |= 0x01; 
				keyports[7] &= 0xfd; keyports[6] &= 0xfd;
#endif
			}
			break;

		case SCANCODE_8:
			if (!ShiftPressed) {
				keyports[4] &= 0xfb;
			}
			else { /* SCANCODE_ASTERISK */
				keyports[4] &= 0xbf; 
			}
			break;

		case SCANCODE_9:
			if (!ShiftPressed) {
				keyports[4] &= 0xfd;
			}
			else { /* SCANCODE_LEFTPAR */
				keyports[4] &= 0xfb;
			}
			break;

		case SCANCODE_0:
			if (!ShiftPressed) {
				keyports[4] &= 0xfe;
			}
			else { /* SCANCODE_RIGHTPAR */
				keyports[4] &= 0xfd;
			}
			break;


		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case SCANCODE_KEYPAD0:
			keyports[5] &= 0x7f;
			break;
		case SCANCODE_KEYPAD1:
			keyports[0] &= 0xdf;
			break;
		case SCANCODE_KEYPAD2:
			keyports[0] &= 0xbf;
			break;
		case SCANCODE_KEYPAD3:
			keyports[0] &= 0x7f;
			break;
		case SCANCODE_KEYPAD4:
			keyports[1] &= 0xdf;
			break;
		case SCANCODE_KEYPAD5:
			keyports[1] &= 0xbf;
			break;
		case SCANCODE_KEYPAD6:
			keyports[1] &= 0x7f;
			break;
		case SCANCODE_KEYPAD7:
			keyports[2] &= 0xdf;
			break;
		case SCANCODE_KEYPAD8:
			keyports[2] &= 0xbf;
			break;
		case SCANCODE_KEYPAD9:
			keyports[2] &= 0x7f;
			break;
		
		/* Finally the <inv> and (c) keys are mapped onto the
		   keypad <enter> and <.> */
		case SCANCODE_KEYPADENTER:
			keyports[7] &= 0x7f;
			break;

		case SCANCODE_KEYPADPERIOD:
			keyports[0] &= 0xfe; keyports[5] &= 0xbf;
			break;

		/* For the shift key we need to set a flag */
		case SCANCODE_LEFTSHIFT: case SCANCODE_RIGHTSHIFT:
			keyports[0] &= 0xfe; 
			ShiftPressed = True;
			break;

		default:
			if (ks < 52) {
				keyports[keytable[ks].port] &= keytable[ks].bit;
			}
			break;
	}
}

/* Process a KeyRelease event */
//static
void process_keyrelease(int ks)
{
	/* First check for the 'special keys' */
	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] &= ~0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] &= ~0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] &= ~0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] &= ~0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] &= ~0x01;

	switch (ks) {


		/* Ordinary SAM keys */
		case SCANCODE_SPACE:
			keyports[7] |= ~0xfe; break;
		case SCANCODE_TAB:
			keyports[3] |= ~0xbf; break;
		case SCANCODE_ENTER:
			keyports[6] |= ~0xfe; break;
		case SCANCODE_LEFTCONTROL: case SCANCODE_RIGHTCONTROL:
			ctrlkey=0; //for handling Ctrl+F1,F2,F3 etc.
		/* <Shift> <Control> is mapped to Caps Lock */
			keyports[8] |= ~0xfe;
			keyports[3] |= ~0x7f;
			break;
		case SCANCODE_LEFTALT:
/*              case SCANCODE_LEFTMETA: ????? */
			keyports[7] |= ~0xfd; break;

		case SCANCODE_RIGHTALT: 
/*              case SCANCODE_RIGHTMETA: case SCANCODE_Mode_switch: ????? */
			keyports[6] |= ~0x7f; break;
		case SCANCODE_BACKSPACE: 
/*              case SCANCODE_DELETE: ????? */
			keyports[4] |= ~0x7f; break;
		case SCANCODE_ESCAPE:
			keyports[3] |= ~0xdf; break;
		case SCANCODE_CURSORBLOCKUP:
			keyports[8] |= ~0xfd; break;
		case SCANCODE_CURSORBLOCKDOWN:
			keyports[8] |= ~0xfb; break;
		case SCANCODE_CURSORBLOCKLEFT:
			keyports[8] |= ~0xf7; break;
		case SCANCODE_CURSORBLOCKRIGHT:
			keyports[8] |= ~0xef; break;

		case SCANCODE_KEYPADDIVIDE:
			keyports[0] |= ~0xfe; 
			keyports[4] |= ~0xdf;
			break;

		case SCANCODE_KEYPADMULTIPLY:
			keyports[0] |= ~0xfe; 
			keyports[4] |= ~0xbf; 
			break;

		case SCANCODE_KEYPADMINUS:
			keyports[4] |= ~0xdf;
			break;

		case SCANCODE_KEYPADPLUS:
			keyports[4] |= ~0xbf;
			break;			

		case SCANCODE_MINUS:
			if (!ShiftPressed)   
				keyports[4] |= ~0xdf;
			else /* SCANCODE_UNDERSCORE */
				keyports[5] |= ~0xdf;
			break;
		case SCANCODE_EQUAL:
			if (!ShiftPressed) {   
				keyports[5] |= ~0xdf;
				keyports[4] |= ~0xbf;
			}
			else { /* SCANCODE_PLUS */
				keyports[0] &= ~0x01; 
				keyports[4] |= ~0xbf;
			}
			break;
		case SCANCODE_SEMICOLON:
			if (!ShiftPressed) {   
				keyports[6] |= ~0xdf;
				keyports[6] |= ~0xbf;
			}
			else { /* SCANCODE_COLON */
                                keyports[0] &= ~0x01;
				keyports[6] |= ~0xbf;
			}
			break;
		case SCANCODE_COMMA:
			if (!ShiftPressed) {
				keyports[7] |= ~0xdf;
				keyports[7] |= ~0xfd;
				keyports[2] |= ~0xfe; 
			}
			else { /* SCANCODE_LESS */
                                keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd;
				keyports[2] |= ~0xfe; 
			}
			break;  
		case SCANCODE_PERIOD:
			if (!ShiftPressed) { 
				keyports[7] |= ~0xbf;
				keyports[7] |= ~0xfd;
				keyports[2] |= ~0xfd;
			}
			else { /* SCANCODE_GREATER */
                                keyports[0] &= ~0x01;
				keyports[7] |= ~0xfd;
				keyports[2] |= ~0xfd;
			}
			break;
		case SCANCODE_SLASH:
			if (!ShiftPressed) {
				keyports[0] |= ~0xfe; 
				keyports[4] |= ~0xdf;
				keyports[7] |= ~0xfd;
				keyports[0] |= ~0xfb;
			}
			else { /* SCANCODE_QUESTION */
                                keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd;
				keyports[0] |= ~0xfb;
			}
			break;
		case SCANCODE_UK_BACKSLASH:
			if (!ShiftPressed) {
				keyports[0] |= ~0xfe; keyports[7] |= ~0x7f;
				keyports[7] |= ~0xfd;
				keyports[4] |= ~0xfd;
			}
			else {  /* SCANCODE_BAR */      
                                keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd;
				keyports[4] |= ~0xfd;
			}
			break;
		case SCANCODE_7:
			if (!ShiftPressed) 
				keyports[4] |= ~0xf7;
			else /* SCANCODE_AMPERSAND */                
				keyports[4] |= ~0xef; 
			break;
		case SCANCODE_BRACKET_LEFT:
			if (!ShiftPressed) {
				keyports[7] |= ~0xfd; keyports[2] |= ~0xf7;
				keyports[7] |= ~0xfd; keyports[1] |= ~0xf7;
			}
			else { /* SCANCODE_BRACELEFT */
				keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd; keyports[1] |= ~0xf7;
			}
			break;
		case SCANCODE_BRACKET_RIGHT:
			if (!ShiftPressed) {
				keyports[7] |= ~0xfd; keyports[2] |= ~0xef;
				keyports[7] |= ~0xfd; keyports[1] |= ~0xef;
			}
			else { /* SCANCODE_BRACERIGHT */
				keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd; keyports[1] |= ~0xef;
			}
			break;
		case SCANCODE_2:
			if (!ShiftPressed) {
				keyports[3] |= ~0xfd;
				keyports[5] |= ~0xbf;				
			}
			else { /* SCANCODE_QUOTEDBL */
#ifdef US_KEYBOARD
				keyports[3] |= ~0xfd;
#else
				keyports[0] &= ~0x01; 
				keyports[5] |= ~0xbf;
#endif
			}                       
			break;

		case SCANCODE_APOSTROPHE:
			if (!ShiftPressed) { 
				keyports[0] |= ~0xfe; 
				keyports[4] |= ~0xf7;
			}
			else { /* SCANCODE_AT */
#ifdef US_KEYBOARD
				keyports[0] &= ~0x01; 
				keyports[5] |= ~0xbf;
#else
				keyports[3] |= ~0xfd;
#endif
			}
			break;

#ifdef US_KEYBOARD
		case SCANCODE_BACKSLASH:
			if (!ShiftPressed) {
				//[\]
				keyports[0] |= ~0xfe;
				keyports[7] |= ~0x7f;
			} else {
				//[|] (Symbol+9)
				keyports[0] &= 0xfe;
				keyports[7] |= ~0xfd;
				keyports[4] |= ~0xfd;
			}
			break;
#else
		case SCANCODE_UK_HASH:   /* NB Not in vgakeyboard.h */
			if (!ShiftPressed) {
				//hash
				keyports[0] |= ~0xfe; 
				keyports[3] |= ~0xfb;
			} else {
				//[~]
				keyports[4] |= ~0xfe;
			}
			break;
#endif

#ifdef US_KEYBOARD
		case SCANCODE_GRAVE:
			if(!ShiftPressed) {
				//[`] = sterling
				keyports[7] |= ~0xfd; 
				keyports[6] |= ~0xfd;
			} else {
				//[~] = tilde
				keyports[4] |= ~0xfe;
				keyports[3] |= ~0xfb;
			}
			break;
#endif

		case SCANCODE_6:
			if (!ShiftPressed) {
				keyports[4] |= ~0xef;
				keyports[7] |= ~0xfd;
				keyports[6] |= ~0xef;
			}
			else { /* SCANCODE_ASCIICIRCUM */
				keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd;
				keyports[6] |= ~0xef;
			}
			break;

		case SCANCODE_3:
			if (!ShiftPressed) {
				keyports[3] |= ~0xfb;
				keyports[7] |= ~0xfd; 
				keyports[6] |= ~0xfd;
			}
			else { /* SCANCODE_STERLING */
#ifdef US_KEYBOARD
				keyports[0] &= ~0x01; 
				keyports[3] |= ~0xfb;
#else
				keyports[0] &= ~0x01; 
				keyports[7] |= ~0xfd; 
				keyports[6] |= ~0xfd;
#endif
			}
			break;

		case SCANCODE_8:
			if (!ShiftPressed) {
				keyports[4] |= ~0xfb;
			}
			else { /* SCANCODE_ASTERISK */
				keyports[4] |= ~0xbf; 
			}
			break;

		case SCANCODE_9:
			if (!ShiftPressed) {
				keyports[4] |= ~0xfd;
			}
			else { /* SCANCODE_LEFTPAR */
				keyports[4] |= ~0xfb;
			}
			break;

		case SCANCODE_0:
			if (!ShiftPressed) {
				keyports[4] |= ~0xfe;
			}
			else { /* SCANCODE_RIGHTPAR */
				keyports[4] |= ~0xfd;
			}
			break;


		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case SCANCODE_KEYPAD0:
			keyports[5] |= ~0x7f;
			break;
		case SCANCODE_KEYPAD1:
			keyports[0] |= ~0xdf;
			break;
		case SCANCODE_KEYPAD2:
			keyports[0] |= ~0xbf;
			break;
		case SCANCODE_KEYPAD3:
			keyports[0] |= ~0x7f;
			break;
		case SCANCODE_KEYPAD4:
			keyports[1] |= ~0xdf;
			break;
		case SCANCODE_KEYPAD5:
			keyports[1] |= ~0xbf;
			break;
		case SCANCODE_KEYPAD6:
			keyports[1] |= ~0x7f;
			break;
		case SCANCODE_KEYPAD7:
			keyports[2] |= ~0xdf;
			break;
		case SCANCODE_KEYPAD8:
			keyports[2] |= ~0xbf;
			break;
		case SCANCODE_KEYPAD9:
			keyports[2] |= ~0x7f;
			break;
		
		/* Finally the <inv> and (c) keys are mapped onto the
		   keypad <enter> and <.> */
		case SCANCODE_KEYPADENTER:
			keyports[7] |= ~0x7f;
			break;
		case SCANCODE_KEYPADPERIOD:
			keyports[0] |= ~0xfe; keyports[5] |= ~0xbf;
			break;

		/* For the shift key we need to reset a flag */
		case SCANCODE_LEFTSHIFT: case SCANCODE_RIGHTSHIFT:
			keyports[0] |= ~0xfe; 
			ShiftPressed = False;
			break;


		default:
			if (ks < 52) {
				keyports[keytable[ks].port] |= ~keytable[ks].bit;
			}
			break;
	}
}

/*
 * Display initialisation routines
 */

/*
 * Initialise the 128 colours required by the SAM
 * Return -1 if colour allocation failed.
 */
int
get_colours(void)
{
	int i;
	int red_index, green_index, blue_index;

	int RGBvals[8] = {0x00, 0x09, 0x12, 0x1b,  
			  0x24, 0x2d, 0x36, 0x3f};


	for (i = 0; i < NCOLOURS; i++) {

		blue_index  = (i & 0x01) << 1| ((i & 0x10) >> 2) | ((i & 0x08) >> 3);                       
		red_index   = (i & 0x02)     | ((i & 0x20) >> 3) | ((i & 0x08) >> 3);                       
		green_index = (i & 0x04) >> 1| ((i & 0x40) >> 4) | ((i & 0x08) >> 3);                       

		VideoDelayedSetPalette(i, RGBvals[red_index],
				     RGBvals[green_index],
				     RGBvals[blue_index]);

		pixels[i] = (long)i;
	}
	VideoLoadPalette();
	return 0;
}


/*
 *  Public functions start here
 */

void
init_display(void)
{
	int i;

	VgaModeAvailable[0] = False;
	VgaModeAvailable[1] = False;

	/* Initialise the keysym -> SAM action translations */

	key_table = (int *)malloc(sizeof(int) * NUM_KEYMAPPINGS);
	for (i = 0; i < NUM_KEYMAPPINGS; i++) {
		key_table[i] = map_tab[i].Scancode;
	}

#ifdef USE_AUDIO_TSR
	/* Audio Initialisation */
	if ((SAARegPtr = AudioInit(GetCfg(audioTSR))) == NULL) {
		simcoupe_mesg(SIMCOUPE_WARN,"Audio failure");
	}
	else {
		OnQuit(AudioClose);
	}
	
#endif		
	/* initialise the graphics code */
	
	if (!VideoInit()) {
		simcoupe_mesg(SIMCOUPE_FATAL, "Cannot initialise Video");	
	}
		
	/* Test to see what vga modes are available */
	Video1x1Mode = VideoFindMode(VSize1x1, HSize1x1);
	Video2x2Mode = VideoFindMode(VSize2x2, HSize2x2);
	
	if (Video1x1Mode != -1) {
		simcoupe_mesg(SIMCOUPE_INFO, "Found 320x200x256 VGA mode");
		VgaModeAvailable[0] = True;
	}
	if (Video2x2Mode != -1) {
		simcoupe_mesg(SIMCOUPE_INFO, "Found 640x480x256 SVGA mode");
		VgaModeAvailable[1] = True;
		simcoupe_mesg(SIMCOUPE_INFO, "Testing for linear addressing");
		if (VideoModeIsLinear(Video2x2Mode)) {
			simcoupe_mesg(SIMCOUPE_INFO,"Linear addressing available");
			Vga2x2LinearAddressing = True;
		}
		else {
			simcoupe_mesg(SIMCOUPE_INFO,"Linear addressing not available - using paging");
			Vga2x2LinearAddressing = False;
		}
				
	}

	/* We must have at least one available video mode to start! */
	if (!(VgaModeAvailable[0] || VgaModeAvailable[1])) {
		simcoupe_mesg(SIMCOUPE_FATAL, "Cannot find a usable video mode");
	}

	flash_phase = !GetCfg(flashing);


	/* Initialise variables used to keep track of line interrupt 
	   clut changes */

	ClutChangeInFrame   = 0;
	BorderChangeInFrame = 0;
	VmprChangeInFrame   = 0;        

        /* Start in SVGA (640x480) mode if one is available */
	if (VgaModeAvailable[1]) {
		ScaleFactor = 2;
	}
	else {
		ScaleFactor = 1;
	}

	init_buffer_mode();
		
	ChangeScaleMode(ScaleFactor);
	line_update = line_update_mode_1;

	/* Initialise the colourmap - NB we have to redo this everytime
	   the vga mode is changed! - here we are just testing the colours
	   are all available */
	
	if (get_colours() == -1) {
		simcoupe_mesg(SIMCOUPE_FATAL, "Colourmap initialisation failed!");
	}

	/* Keyboard initialisation */
	
	if (KeyboardInit(SamKeyboardHandler)) {
		simcoupe_mesg(SIMCOUPE_FATAL, "Can't initialise keyboard");
	}       	

	/* No keys initially being pressed */
	for (i = 0; i < 9; i++)
		keyports[i] = 0xff;
	/* Keyboard starts off in SAM mode by default */
	SpectrumKeyboard = 0;

#ifdef JOY
	/* Linux analogue joystick initialisation */
	if (GetCfg(use_js) && joystick_init() != 0) { 
		simcoupe_mesg(SIMCOUPE_WARN, "joystick initialisation failed");
		SetCfg(use_js, 0);
	}
#endif

	/* Mouse initialisation */

	if (MouseInit(SamMouseHandler)) {
		simcoupe_mesg(SIMCOUPE_WARN,"Can't initialise mouse");
	}       
	sam_mouse_init();

#if 0
    /* Parallel printer initialisation */
    PrinterInit();
#endif

	OnQuit(closedown);
}


/* Check & process any incoming events */
void 
check_events(void)
{
	static FirstTime = True;

	/* If this is the first call to check_events(), after */
	/* the first emulation frame, then use the UI to put  */
	/* up a splash banner.				      */

	if (FirstTime) {
		FirstTime = False;
		ui_banner();
		return;
	}
	
	/* Actually all we do is to call the mouse and
	   keyboard update routines - which in turn call the
	   appropriate update routines */

	if (SAMPointerMode) {
		sam_mouse_xy_write(0, 0);
		if (UiActive) ui_set_pointer_position(0, 0);
	}
	MouseUpdate();
	EmulatorControl();

	/* If the UI is active call its event handler*/
	
	if (UiActive) ui_event();
}


/* Keyboard handler - control is passed to here if a keyboard event is  */
/* trapped by the interrupt handler - This is unique to the DOS version */

static void
SamKeyboardHandler(int Scancode, int EventType)
{
	switch (EventType) {
	
		case KEY_EVENTPRESS:
		    if ((Scancode >= SCANCODE_F1 && Scancode <= SCANCODE_F10) ||
			(Scancode == SCANCODE_F11) || (Scancode == SCANCODE_F12)) {
	     		EmulatorControlEvent = Scancode;
		    }
		    else {
	
			if (!SpectrumKeyboard)
				process_keypress(Scancode);
			else
				process_spectrum_keypress(Scancode);
			break;
		    }
		case KEY_EVENTRELEASE:
			if (!SpectrumKeyboard)
				process_keyrelease(Scancode);
			else
				process_spectrum_keyrelease(Scancode);
			break;
		default:
			simcoupe_mesg(SIMCOUPE_ERR,"Unknown Key event type");   

	} /* end switch */

}


/* Emulator control keys - dealt with once per frame, and flagged */
/* by interrupt driven keyboard handler				  */

//#define FK_HELP SCANCODE_F1
#define FK_MENU SCANCODE_F2
#define FK_DISK1 SCANCODE_F3
#define FK_DISK2 SCANCODE_F4

#define FK_VIDEOMODE SCANCODE_F5
#define FK_MOUSEMODE SCANCODE_F6
#define FK_KEYMODE SCANCODE_F7

#define FK_NMI SCANCODE_F11
#define FK_RESET SCANCODE_F12


static void
EmulatorControl(void)
{
	int i;
	
	if (EmulatorControlEvent) {
	    switch (EmulatorControlEvent) {
		/* Emulator control keys */

		case FK_KEYMODE:
			SpectrumKeyboard = !SpectrumKeyboard;
			break;

		case FK_MOUSEMODE:
			//Ctrl+F3 switches between modes
			if(ctrlkey) {
				if(SAMPointerMode) SAMPointerMode^=1;
				break;
			}

			//F3 turns mouse on/off
			if (!SAMPointerMode) {
				SAMPointerMode=1;
				sam_mouse_button_write(PointerButtonsOff);
				sam_mouse_xy_write(0, 0);	
				if (UiActive) {
				    ui_set_button_state(PointerButtonsOff);
				    ui_set_pointer_position(0, 0);
				    ui_pointer_draw();
				}
			} else {
				SAMPointerMode=0;
				ui_pointer_erase();
			}
			break;

		case FK_VIDEOMODE:
			if (VgaModeAvailable[(ScaleFactor == 1) ? 1 : 0]) {
				ScaleFactor = (ScaleFactor == 1) ? 2 : 1;
				ChangeScaleMode(ScaleFactor);

				/* Call vmpr to reset the screen write pointers and
				   fetch the screen from memory */
				call_vmpr(vmpr);
			}       
			break;

		case FK_NMI:
			InterruptType = Z80_nmi;
			break;
			
		case FK_DISK1:
			if (!UiActive) ui_main(UI_DISK1);
			break;

		case FK_DISK2:
			if (!UiActive) ui_main(UI_DISK2);
			break;

		case FK_MENU:
			if (!UiActive) ui_main(UI_MENU);
			break;

		case FK_RESET:
			//Ctrl+F12 = exit SimCoupe
			if(ctrlkey) Quit(0);

			//else reset emulator
			InterruptType = Z80_reset;
			for (i = 0; i < 9; i++)
				keyports[i] = 0xff;
			break;

	    }
	}
	EmulatorControlEvent = False;
}


/* Mouse handler - control is passed to here if an event is */
/* pending when MouseUpdate() is called from check_events() */

static void
SamMouseHandler(int Button, int dx, int dy)
{
	int SAMButtons;

	static int OldButton = 0xff;

	/* process a pointer move event */
	if (SAMPointerMode && (dx || dy)) {
		dx = (dx*SAMPointerMode);
		dy = - dy;
		sam_mouse_xy_write(dx, dy);
		/* Update the pointer position used for the UI if active */
		
		if (UiActive)
			ui_set_pointer_position(dx, -dy);
	}

	if (OldButton != Button) {

		OldButton = Button;

		if (Button & MOUSE_LEFTBUTTON)
			SAMButtons &= ~SAM_BUTTON_1;
		else
			SAMButtons |=  SAM_BUTTON_1;
	
		if (Button & MOUSE_MIDDLEBUTTON)
			SAMButtons &= ~SAM_BUTTON_2;
		else
			SAMButtons |=  SAM_BUTTON_2;
	
		if (Button & MOUSE_RIGHTBUTTON)
			SAMButtons &= ~SAM_BUTTON_3;
		else
			SAMButtons |=  SAM_BUTTON_3;

		sam_mouse_button_write(SAMButtons);

		/* Update the UI Buttons state if active */
		if (UiActive)
			ui_set_button_state(SAMButtons);

	}       
}       
